@@ -35,7 +35,7 @@ ZTFIRSA_BASE = ["/ref/","/sci/","/raw/","/cal/"]
 from .io import LOCALSOURCE, CCIN2P3_SOURCE
 DEFAULT_SOURCE = DATA_BASEURL
 
-FILTERS = {"zg":1,"zr":2,"zi":3, "OO":None}
+FILTERS = {"zg":1,"zr":2,"zi":3, "OO":None, "bi":None}
 
 
 
@@ -44,15 +44,28 @@ def get_rawfile_of_filename(filename, source="irsa"):
     """ """
 
     parsing = parse_filename(filename)
-    kind = parsing["kind"]
-    url = filename_to_url(filename, kind=kind)
-
     func_argument = ["year", "month", "day", "fracday", "filtercode",
                      "imgtypecode", "paddedfield"]
         
     prop = {k:parsing[k] for k in func_argument}
     prop["paddedccdid"] = f"{parsing['ccdid']:02d}"
     return raw_path(source=source, **prop)
+
+def get_scifile_of_filename(filename, qid=None, source="irsa"):
+    """ """
+    parsing = parse_filename(filename)
+    func_argument = ["year", "month", "day", "fracday", "paddedfield","filtercode",
+                     "paddedccdid", "qid"]
+    # work all the time
+    prop = {k:str(parsing[k]) for k in func_argument if k not in ["qid", "paddedccdid"]}
+    prop["paddedccdid"] = f"{parsing['ccdid']:02d}"
+    if parsing["kind"] == "raw":
+        if qid is None:
+            return [science_path(qid=str(i), **prop, source=source)
+                        for i in range(1,5)]
+        return science_path(qid=str(qid), **prop, source=source)
+    return science_path(qid=parsing["qid"], **prop, source=source)
+    
         
 
 # ================== #
@@ -86,6 +99,7 @@ def filename_to_kind(filename):
 
 def parse_calfilename(filename):
     """ """
+    from .fields import ccdid_qid_to_rcid    
     _, fileday, filtercode, ccd_, qid_, suffix = os.path.basename(filename).split("_")
     year,month, day = fileday[:4],fileday[4:6], fileday[6:]
     ccdid = int(ccd_.replace("c",""))
@@ -98,7 +112,7 @@ def parse_calfilename(filename):
             "qid":qid,
             "rcid":ccdid_qid_to_rcid(ccdid,qid),
             "filtercode":filtercode,
-            "filterid":FILTERS[filtercode],
+            "filterid":FILTERS.get(filtercode,None),
             "kind":"cal",
             "type":suffix.split(".")[0]}
 
@@ -117,7 +131,7 @@ def parse_rawfilename(filename):
             "field":int(paddedfield),
             "ccdid":ccdid,
             "filtercode":filtercode,
-            "filterid":FILTERS[filtercode],
+            "filterid":FILTERS.get(filtercode, None),
             "kind":"raw",
             "suffix":suffix_}
 
@@ -143,7 +157,7 @@ def parse_scifilename(filename):
             "qid":qid,
             "rcid":ccdid_qid_to_rcid(ccdid,qid),
             "filtercode":filtercode,
-            "filterid":FILTERS[filtercode],
+            "filterid":FILTERS.get(filtercode, None),
             "kind":"sci",
             "suffix":"_".join(suffix_)}
 
