@@ -6,19 +6,21 @@ import os, sys
 import pandas
 import numpy as np
 import warnings
-from pandas import read_csv
+
 from astropy import units, coordinates, time
 import matplotlib.pyplot as mpl
 from matplotlib.patches import Polygon
 
 
+from .io import _SOURCEDIR
 
-
-_FIELD_SOURCE = os.path.dirname(os.path.realpath(__file__))+"/data/ztf_fields.txt"
-FIELD_DATAFRAME = read_csv(_FIELD_SOURCE, index_col="ID")
+_FIELD_SOURCE = os.path.join(_SOURCEDIR, "data/ztf_fields.txt")
+FIELD_DATAFRAME = pandas.read_csv(_FIELD_SOURCE, index_col="ID")
 FIELDSNAMES = FIELD_DATAFRAME.index.values
 
-_CCD_COORDS  = read_csv(os.path.dirname(os.path.realpath(__file__))+"/data/ztf_ccd_layout.tbl").rename(columns={"CCD ":"CCD"}) # corner of each CCDS
+
+_CCD_COORDS  = pandas.read_csv( os.path.join(_SOURCEDIR, "data/ztf_ccd_layout.tbl")).rename(columns={"CCD ":"CCD"}) # corner of each CCDS
+_QUAD_COORDS  = pandas.read_csv( os.path.join(_SOURCEDIR,"data/ztf_ccd_quad_layout.tbl"))#.rename(columns={"CCD ":"CCD"}) # corner of each CCDS
 
 FIELD_COLOR = {1: "C2", 2: "C3", 3:"C1"}
 FIELD_CMAP = {1: mpl.cm.Greens, 2:mpl.cm.Reds, 3:mpl.cm.Oranges}
@@ -26,14 +28,16 @@ FIELDNAME_COLOR = {"zg": "C2", "zr":"C3", "zi":"C1"}
 
 _PLOTORIGIN = 180
 
-def _load_fields_geoserie_(inclccd=False):
+def _load_fields_geoserie_(inclquad=False, inclccd=False):
     """ Loads the FIELDS_GEOSERIE global variable 
     = Internal Tools =
     """
-    if not inclccd:
-        global FIELDS_GEOSERIE
-    else:
+    if inclquad:
+        global FIELD_QUADS_GEOSERIE
+    elif inclccd:
         global FIELD_CCDS_GEOSERIE
+    else:
+        global FIELDS_GEOSERIE
     
     try:
         from geopandas import geoseries
@@ -45,25 +49,35 @@ def _load_fields_geoserie_(inclccd=False):
             FIELD_CCDS_GEOSERIE = None 
         return
     
-    field_verts = get_field_vertices(fieldid=FIELDSNAMES, inclccd=inclccd, asdict=True, aspolygon=True)
-    if not inclccd:
-        FIELDS_GEOSERIE = geoseries.GeoSeries(field_verts)
-    else:
+    field_verts = get_field_vertices(fieldid=FIELDSNAMES, inclquad=inclquad, inclccd=inclccd, as_dict=True, as_polygon=True)
+    if inclquad:
+        FIELD_QUADS_GEOSERIE = geoseries.GeoSeries(field_verts)
+    elif inclccd:
         FIELD_CCDS_GEOSERIE = geoseries.GeoSeries(field_verts)
+    else:
+        FIELDS_GEOSERIE = geoseries.GeoSeries(field_verts)
 
     
-def get_fields_geoserie(inclccd=False):
+def get_fields_geoserie(inclquad=False, inclccd=False):
     """ returns the global variable FIELDS_GEOSERIE, creates it if necessary """
-    if not inclccd:
-        if not hasattr(sys.modules[__name__], 'FIELDS_GEOSERIE'):
-            _load_fields_geoserie_(inclccd=False)
+    if inclquad:
+        if not hasattr(sys.modules[__name__], 'FIELD_QUADS_GEOSERIE'):
+            _load_fields_geoserie_(inclccd=False, inclquad=True)
         
-        return FIELDS_GEOSERIE
-    else:
+        return FIELD_QUADS_GEOSERIE
+    
+    elif inclccd:
         if not hasattr(sys.modules[__name__], 'FIELD_CCDS_GEOSERIE'):
             _load_fields_geoserie_(inclccd=True)
         
         return FIELD_CCDS_GEOSERIE
+    
+    else:
+        if not hasattr(sys.modules[__name__], 'FIELDS_GEOSERIE'):
+            _load_fields_geoserie_(inclccd=False)
+        
+        return FIELDS_GEOSERIE
+        
 
 # ------------------ #
 #                    #
@@ -164,8 +178,72 @@ def get_field_ccd_qid(ra, dec):
         d_[int(field)] = {"ccd":int(ccd), "qid":int(qid), "rcid":ccdid_qid_to_rcid(int(ccd), int(qid))}
         
     return d_
+
+
+def spatialjoin_radec_to_fields(radec, fields, how="inner", predicate="intersects",
+                                    index_radec="index_radec",**kwargs):
+    """ 
+    radec: DataFrame or 2d-array 
+        coordinates of the points. 
+        - DataFrame: must have the "ra" and "dec" columns. 
+            This will use the DataFrame's index are data index.
+        - 2d array (shape N,2): returned index will be 'range(len(ra))'
     
-def get_fields_containing_target(ra, dec, inclccd=False):
+    fields : [geopandas.geoserie, geopandas.geodataframe or  dict]
+        fields contains the fieldid and fields shapes. Several forms are accepted:
+        - dict: {fieldid: 2d-array, fieldid: 2d-array ...}
+            here, the 2d-array are the field's vertices.
+
+        - geoserie: geopandas.GeoSeries with index as fieldid and geometry as field's vertices.
+            
+        - geodataframe: geopandas.GeoDataFrame with the 'fieldid' column and geometry as field's vertices.
+
+    Returns
+    -------
+    GeoDataFrame (geometry.sjoin result )
+    """
+    import geopandas
+    
+    # -------- #
+    #  Coords  #
+    # -------- #
+    if type(radec) in [np.ndarray, list, tuple]:
+        if (inshape:=np.shape(radec))[-1] != 2:
+            raise ValueError(f"shape of radec must be (N, 2), {inshape} given.")
+        
+        radec = pandas.DataFrame(radec, columns=["ra","dec"])
+
+    # Points to be considered
+    geoarray = geopandas.points_from_xy(*radec[["ra","dec"]].values.T)
+    geopoints = geopandas.GeoDataFrame({index_radec:radec.index}, geometry=geoarray)
+    
+    # -------- #
+    # Fields   #
+    # -------- #
+    # goes from dict to geoseries (more natural) 
+    if type(fields) is dict:
+        values = fields.values()
+        indexes = fields.keys()
+        # dict of array goes to shapely.Geometry as expected by geopandas
+        if type(values.__iter__().__next__()) in [np.ndarray, list, tuple]:
+            from shapely import geometry
+            values = [geometry.Polygon(v) for v in values]
+        
+        fields = geopandas.GeoSeries(values,  index = indexes)
+            
+    if type(fields) is geopandas.geoseries.GeoSeries:
+        fields = geopandas.GeoDataFrame({"fieldid":fields.index},
+                                        geometry=fields.values)
+    elif type(fields) is not geopandas.geodataframe.GeoDataFrame:
+        raise ValueError("cannot parse the format of the input 'fields' variable. Should be dict, GeoSeries or GeoPandas")
+
+    # -------- #
+    # Joining  #
+    # -------- #
+    return geopoints.sjoin(fields,  how="inner", predicate="intersects", **kwargs)
+
+
+def get_fields_containing_target(ra, dec, inclccd=False, buffer=None):
     """ return the list of fields into which the position ra, dec is. 
     Remark that this is based on predefined field positions. 
     Hence, small attrition could affect this.
@@ -179,6 +257,10 @@ def get_fields_containing_target(ra, dec, inclccd=False):
         do you want the details of the CCD id on top of the field
         format: "field_ccdid"
 
+    buffer: [float] -optional-
+        buffer the polygon (fields or ccds). 
+        In degree (unit of the polygons). The inter-ccd gap typically is 0.3 deg
+
     Returns
     -------
     list (all the field ID that contain the given ra,dec coordinates)
@@ -195,11 +277,15 @@ def get_fields_containing_target(ra, dec, inclccd=False):
         warnings.warn("get_fields_containing_target would be much faster if you install geopandas (pip install geopandas)")
         return [f for f in FIELDSNAMES
                 if geometry.Polygon( get_field_vertices(f)[0]).contains(coordpoint)]
-    
-    return fields_geoserie.index[ fields_geoserie.contains(coordpoint) ]
+    if buffer is None:
+        return fields_geoserie.index[ fields_geoserie.contains(coordpoint) ]
+    return fields_geoserie.index[ fields_geoserie.buffer(buffer).contains(coordpoint) ]
 
 
-def get_field_vertices(fieldid=None, inclccd=False, ccd=None, asdict=False, aspolygon=False,
+def get_field_vertices(fieldid=None,
+                           inclquad=False, qid=None,
+                           inclccd=False, ccd=None,
+                           as_dict=False, as_polygon=False,
                         squeeze=True):
     """ Get the fields countours 
     
@@ -211,68 +297,74 @@ def get_field_vertices(fieldid=None, inclccd=False, ccd=None, asdict=False, aspo
 
     // output format
 
-    asdict: [bool] -optional-
-        Do you want to result as a list (asdict=False) following the input's fieldid sorting
+    as_dict: [bool] -optional-
+        Do you want to result as a list (as_dict=False) following the input's fieldid sorting
         or do you want a dict ({fieldid_: verts_ ... })
 
-    aspolygon: [bool] -optional-
-        Do you want the vertices as 2d-array (aspolygon=False) or as shapely Geometries (True)
+    as_polygon: [bool] -optional-
+        Do you want the vertices as 2d-array (as_polygon=False) or as shapely Geometries (True)
 
     squeeze: [bool] -optional-
         Should unnecessary dimension be removed ?
-        if asdict is True:
+        if as_dict is True:
             if squeeze: {fid_ccdid: }
             if not squeeze: {fid:{ccdid: }}
             = if not inclccd, squeez ignored =
-        if not asdict:
+        if not as_dict:
            using np.squeeze, basically doing [[]]->[]
     Returns
     -------
-    list of dict (see asdict)
+    list of dict (see as_dict)
     """
     if fieldid is None:
         fieldid = FIELDSNAMES
         
     if inclccd and ccd is None:
         ccd = np.arange(1,17)
+        
+    if inclquad and qid is None:
+        qid = np.arange(0,64)
 
     # - Actual calculation
     rafields, decfields  = get_field_centroid( np.asarray(np.atleast_1d(fieldid), dtype="int") ).T
-    fields_verts = get_corners(rafields, decfields, inclccd=inclccd, ccd=ccd,
-                                       inrad=False, squeeze=False)
+    fields_verts = get_corners(rafields, decfields,
+                                inclquad=inclquad, qid=qid,
+                                inclccd=inclccd, ccd=ccd,
+                                inrad=False, squeeze=False)
 
     # ----------- #
     #  Format     #
     # ----------- #    
-    if aspolygon:
+    if as_polygon:
         try:
             from shapely import geometry
-            fields_countours = [[geometry.Polygon(fields_verts[f_][c_])
-                                         for c_,ccd_ in enumerate(np.atleast_1d(ccd))]
-                                         for f_,field_ in enumerate(np.atleast_1d(fieldid))]
+            shape = np.shape(fields_verts)
+            shape_flat = tuple([np.prod(shape[:-2])] + list(shape[-2:]))
+            fields_countours = np.reshape([geometry.Polygon(f_) for f_ in fields_verts.reshape(shape_flat)], shape[:-2])
         except ImportError:
-            warnings.warn("You do not have shapely, Please run pip install shapely. 'aspolygon' set to False")
+            warnings.warn("You do not have shapely, Please run pip install shapely. 'as_polygon' set to False")
     else:
         fields_countours = fields_verts
 
     # ----------- #
     #  Output     #
     # ----------- #    
-    if not asdict:
+    if not as_dict:
         return fields_countours if not squeeze else np.squeeze(fields_countours)
 
     # full camera dict
-    if not inclccd:
-        return {i:k[0] for i,k in zip(fieldid,fields_countours)}
-    
-    # ccd dict
-    if squeeze:
-        return {f"{field_}_{ccd_}":fields_countours[f_][c_]
-                    for c_,ccd_ in enumerate(np.atleast_1d(ccd)) for f_,field_ in enumerate(np.atleast_1d(fieldid))}
-    else:
-        return {field_:{ccd_:fields_countours[f_][c_] for c_,ccd_ in enumerate(np.atleast_1d(ccd))}
+    if inclquad:
+        return {f"{field_}_{qid_}": fields_countours[f_][c_]
+                    for c_,qid_ in enumerate(np.atleast_1d(qid))
                     for f_,field_ in enumerate(np.atleast_1d(fieldid))}
-
+    
+    elif inclccd:
+        return {f"{field_}_{ccd_}": fields_countours[f_][c_]
+                    for c_,ccd_ in enumerate(np.atleast_1d(ccd))
+                    for f_,field_ in enumerate(np.atleast_1d(fieldid))}
+    else:
+        return {i:k[0] for i,k in zip(fieldid, fields_countours)}
+    
 def get_field_centroid(fieldid, system="radec"):
     """ Returns the central coordinate [RA,Dec] or  of the given field 
 
@@ -298,24 +390,35 @@ def get_field_centroid(fieldid, system="radec"):
         syst = ["Ecl Long","Ecl Lat"]
     else:
         raise ValueError("unknown coordinate system %s select among: [radec / galactic / ecliptic]"%system)
+    
     fieldid = np.atleast_1d(fieldid)
     radec = np.asarray(FIELD_DATAFRAME[np.in1d(FIELD_DATAFRAME.index, fieldid)][syst].values)
     
     return radec
 
-def get_corners(ra_field, dec_field, inclccd=False, ccd=None, steps=5, squeeze=True, inrad=False):
+def get_corners(ra_field, dec_field, inclquad=False, inclccd=False,
+                    qid=None, ccd=None, steps=5, squeeze=True, inrad=False):
     """ """
     from .utils.tools import rot_xz_sph, _DEG2RA
-    
-    if not inclccd:
-        upper_left_corner = _CCD_COORDS.max()
-        lower_right_corner = _CCD_COORDS.min()
-    elif ccd is None:
+
+    if inclquad or qid is not None:
+        upper_left_corner = _QUAD_COORDS.groupby("Quad").max()
+        lower_right_corner = _QUAD_COORDS.groupby("Quad").min()
+        if qid is not None:
+            upper_left_corner = upper_left_corner.loc[qid]
+            lower_right_corner = lower_right_corner.loc[qid]
+    # CCD
+    elif inclccd or ccd is not None:
         upper_left_corner = _CCD_COORDS.groupby("CCD").max()
         lower_right_corner = _CCD_COORDS.groupby("CCD").min()
+        if ccd is not None:
+            upper_left_corner = upper_left_corner.loc[ccd]
+            lower_right_corner = lower_right_corner.loc[ccd]
+    # Focal Plane
     else:
-        upper_left_corner = _CCD_COORDS.groupby("CCD").max().loc[ccd]
-        lower_right_corner = _CCD_COORDS.groupby("CCD").min().loc[ccd]
+        upper_left_corner = _CCD_COORDS.max()
+        lower_right_corner = _CCD_COORDS.min()
+
         
     ewmin = -np.atleast_1d(upper_left_corner["EW"])
     nsmax = np.atleast_1d(upper_left_corner["NS"])
@@ -554,7 +657,7 @@ def show_gri_fields(fieldsg=None, fieldsr=None, fieldsi=None,
                             show_ztf_fields=show_ztf_fields,
                             show_mw=show_mw, mw_b=mw_b, mw_prop=mw_prop,
                             **{**prop,**patch})
-    return _
+    return ax[0].figure
 
 def _get_gri_axes_(alignment="classic", title=None, titlefontsize="large", projection="hammer",
                    labelsize="x-small", labelcolor="0.7",  clabelsize="x-small", clabelcolor="k",
